#define WAF2_LOGGING_ENABLE

#include "RequestFlow.h"
#include "TcpFlow.h"
#include "HtsenseUtil.h"
#include "waf2_logging.h"
#include <cstdlib>
#include <string.h>
#include <strings.h>

void RequestFlow::emit_headers() {
	m_tcpFlow->requestStartedCb(method, method_len, path, path_len, minor_version, tv_sec, tv_usec);

	for (size_t i = 0; i != num_headers; ++i) {
		const char *name = headers[i].name;
		const char *value = headers[i].value;
		int name_len = (int)headers[i].name_len;
		int value_len = (int)headers[i].value_len;

		// Name can be NULL if this is "line continuation".
		// In this case, we should append to previous value.
		if (name != NULL) {
			// name:value pair (the value may be incomplete)
			m_tcpFlow->requestHeaderCb(name, name_len, value, value_len);  // TODO:: very long headers can appear cut in audit log (need to fix!)
		}
		else {
			// continuation of last value
			waf2_log(WAF2_LOG_TRACE, "\t\t%.*s'", value_len, value);
		}
	}

	m_tcpFlow->requestEndHeadersCb();
}

void RequestFlow::emit_body_start() {
	m_tcpFlow->requestBodyStartCb();
}

void RequestFlow::emit_body_chunk(const char *data, size_t data_len, bool last) {
	if (data_len) {
		m_tcpFlow->requestBodyChunkCb(data, data_len);
	}

	if (last) {
		m_tcpFlow->requestEndBodyCb();
	}
}

void RequestFlow::emit_request_end() {
	m_tcpFlow->requestFinishedCb();
}

int RequestFlow::push(const char *buf, size_t buf_len, time_t tv_sec, suseconds_t tv_usec) {
	if (m_total_bytes == 0) {
		// record timestamp of first request packet
		this->tv_sec = tv_sec;
		this->tv_usec = tv_usec;
	}

	m_total_bytes += buf_len;
	m_buffer += std::string(buf, buf_len);
	int bytes_processed = 0;

	while (m_buffer.length() > 0) {
		if (m_state == STATE_READY) {
			m_state = STATE_HEADERS;
		}

		if (m_state == STATE_HEADERS) {
			// Try to parse HTTP request headers
			waf2_log(WAF2_LOG_TRACE, "Parsing request headers...");
			num_headers = sizeof(headers) / sizeof(headers[0]);
			int hdrs_size = phr_parse_request(m_buffer.data(), m_buffer.length(), &method, &method_len,
				&path, &path_len, &minor_version, headers, &num_headers, m_lastBufferLength);
			m_lastBufferLength = m_buffer.length();

			if (hdrs_size > 0) {
				// Succesfully parsed request headers
				waf2_log(WAF2_LOG_TRACE, "Request headers were %d bytes long.", hdrs_size);

				// Look in headers for info required for HTTP state-machine handling
				m_content_length = 0;
				bool is_chunked = false;
				for (size_t i = 0; i != num_headers; ++i) {
					const char *name = headers[i].name;
					const char *value = headers[i].value;
					int name_len = (int)headers[i].name_len;
					int value_len = (int)headers[i].value_len;

					if (0 == strncasecmp(name, "Content-Length", name_len)) {
						// TODO:: theoretically, value can be negative, in which case m_content_length will be huge value.
						// We have to somehow validate such things (maybe not use atoi at all! For security's sake!
						m_content_length = atoi(value);
					}

					if (0 == strncasecmp(name, "Transfer-Encoding", name_len) && 0 == strncasecmp(value, "chunked", value_len)) {
						is_chunked = true;
						m_content_length = 0; // by RFC - content-length MUST BE IGNORED when using non-"identity" transfer enccoding.
						break;
					}
				}

				// Emit parsed request headers before freeing the buffer
				emit_headers();

				// Remove headers (already parsed) from the buffer
				m_buffer.erase(0, hdrs_size);
				bytes_processed += hdrs_size;

				// Switch HTTP state based on information from headers
				if (m_content_length > 0) {
					waf2_log(WAF2_LOG_TRACE, "Switching HTTP parser state to STATE_BODY.");
					m_state = STATE_BODY;
					emit_body_start();
				}
				else if (is_chunked) {
					waf2_log(WAF2_LOG_TRACE, "Switching HTTP parser state to STATE_CHUNKED_BODY.");
					m_state = STATE_CHUNKED_BODY;
					emit_body_start();

					// zero-clear chunked data decoder state structure
					memset((void*)&m_chunk_decoder, 0, sizeof(m_chunk_decoder));

					// set consume_trailer to 1 to discard the trailing header, or the application
					//should call phr_parse_headers to parse the trailing header.
					m_chunk_decoder.consume_trailer = 1;
				}
				else {
					// this request is finished (no body). Get ready to handle next one.
					m_state = STATE_READY;
					emit_request_end();
				}
			}
			else if (hdrs_size == -1) {
				waf2_log(WAF2_LOG_ERROR, "Request headers parse error.");
				m_state == STATE_ERROR;
				return -1;
			}
			else {
				assert(hdrs_size == -2);
				waf2_log(WAF2_LOG_TRACE, "Request headers incomplete. Will restart next time when more data is available.");
				return -2;
			}
		}
		else if (m_state == STATE_BODY) {
			assert(m_content_length > 0);

			if (m_buffer.length() >= m_content_length) {
				emit_body_chunk(m_buffer.data(), m_content_length, true); // last
				m_buffer.erase(0, m_content_length);
				bytes_processed += m_content_length;
				waf2_log(WAF2_LOG_TRACE, "Switching HTTP Request parser state back to STATE_READY.");
				m_state = STATE_READY;
				emit_request_end();
			}
			else {
				waf2_log(WAF2_LOG_TRACE, "Request body incomplete. Will restart next time when more data is available.");
				return -2;
			}
		}
		else if (m_state == STATE_CHUNKED_BODY) {
			waf2_log(WAF2_LOG_TRACE, "Parsing request chunked body.");

			size_t bytes_decoded = m_buffer.length();
			int pret = phr_decode_chunked(&m_chunk_decoder, (char*)m_buffer.data(), &bytes_decoded);
			waf2_log(WAF2_LOG_TRACE, "phr_decode_chunked returned %d, bytes_decoded=%lu", pret, bytes_decoded);

			if (pret == -1) {
				waf2_log(WAF2_LOG_ERROR, "Failed to decode request body chunk");
				m_state == STATE_ERROR;
				return -1;
			}

			// phr_decode_chunked() moves some data within m_buffer effectively reducing data length
			// but m_buffer string is, obviously, not aware of this change in content. Here we update
			// m_buffer to correspond to the new actual length.
			m_buffer.resize(bytes_decoded);

			if (pret == -2) {
				// Avoid emitting 0-length data unnecessarily
				if (bytes_decoded > 0) {
					emit_body_chunk(m_buffer.data(), bytes_decoded, false);
					m_buffer.erase(0, bytes_decoded);
					bytes_processed += bytes_decoded;
				}

				waf2_log(WAF2_LOG_TRACE, "Request body chunk incomplete. Will restart next time when more data is available.");
				return -2;
			}
			else {
				// Here always emit (even if length 0) because we should tell emitter that data ended.
				emit_body_chunk(m_buffer.data(), bytes_decoded, true); // last
				m_buffer.erase(0, bytes_decoded);
				bytes_processed += bytes_decoded;
				waf2_log(WAF2_LOG_TRACE, "Switching HTTP Request parser state back to STATE_READY.");
				m_state = STATE_READY;
				emit_request_end();
			}
		}
	}

	return bytes_processed;
}
