#define WAF2_LOGGING_ENABLE

#include "TcpFlow.h"
#include "HtsenseUtil.h"
#include "waf2_logging.h"
#include <arpa/inet.h>
#include <uuid/uuid.h>
#include <string.h>
#include <boost/smart_ptr/make_shared.hpp>

TcpFlow::TcpFlow(const struct tuple4 &tuple, time_t tv_sec, suseconds_t tv_usec, bool logResponses, bool logResponseBody)
:tuple(tuple), conn_est_tv_sec(tv_sec), conn_est_tv_usec(tv_usec), requestFlow(this), responseFlow(this), m_logResponses(logResponses), m_logResponseBody(logResponseBody) {
}

std::string TcpFlow::name() const {
	return std::string("TCP FLOW ") + " : " + tuple4_to_str(tuple);
}

// Request handling
void TcpFlow::requestStartedCb(const char *method, int method_len, const char *path, int path_len, int http_minor_version, time_t request_tv_sec, suseconds_t request_tv_usec) {
	m_auditLogQueue.push(boost::make_shared<TransactionLogger>());
	m_auditLogQueue.front()->start_transaction();
	waf2_log(WAF2_LOG_TRACE, "HTTP Pipeline starting transaction. Pipeline size now: %lu", m_auditLogQueue.size());

	waf2_log(WAF2_LOG_TRACE, "-- Request parsed:");
	waf2_log(WAF2_LOG_TRACE, "\tRequest method: %.*s ", method_len, method);
	waf2_log(WAF2_LOG_TRACE, "\tRequest path: %.*s", path_len, path);
	waf2_log(WAF2_LOG_TRACE, "\tRequest HTTP version: 1.%d", http_minor_version);

#define int_ntoa(x)     inet_ntoa(*((struct in_addr *)&x))
	char local_addr[64];
	strcpy (local_addr, int_ntoa (tuple.daddr));
	int local_port = tuple.dest;

	char remote_addr[64];
	strcpy (remote_addr, int_ntoa (tuple.saddr));
	int remote_port = tuple.source;

	char conn_est_time[64];
	snprintf(conn_est_time, sizeof(conn_est_time), "%lu.%lu", conn_est_tv_sec, conn_est_tv_usec);

	char request_time[64];
	snprintf(request_time, sizeof(request_time), "%lu.%lu", request_tv_sec, request_tv_usec);

	// generate unique UUID of this specific request/response transaction
	uuid_t transaction_uuid;
	uuid_generate(transaction_uuid);
	char transaction_uuid_str[37];
	uuid_unparse_lower(transaction_uuid, transaction_uuid_str);

	m_auditLogQueue.front()->emit_transaction_info(conn_est_time, transaction_uuid_str, remote_addr, remote_port, local_addr, local_port);
	m_auditLogQueue.front()->start_request(std::string(method, method_len).c_str(), (std::string(path, path_len) + " HTTP/1." + (http_minor_version ? "1" : "0")).c_str(), request_time);

	waf2_log(WAF2_LOG_TRACE, "Headers:");
	m_auditLogQueue.front()->start_request_hdrs();
}

void TcpFlow::requestHeaderCb(const char *name, int name_len, const char *value, int value_len) {
	waf2_log(WAF2_LOG_TRACE, "\t|%.*s: %.*s'", name_len, name, value_len, value);
	m_auditLogQueue.front()->add_request_hdr(name, name_len, value, value_len);
}

void TcpFlow::requestEndHeadersCb() {
	m_auditLogQueue.front()->end_request_hdrs();
	waf2_log(WAF2_LOG_TRACE, "-------");
}

void TcpFlow::requestBodyStartCb() {
	m_auditLogQueue.front()->start_request_body();
}

void TcpFlow::requestBodyChunkCb(const char *data, int data_len) {
	if (waf2_log_level() == WAF2_LOG_TRACE) {
		waf2_log(WAF2_LOG_TRACE, "REQUEST BODY_CHUNK:");
		util_hex_dump(data, data_len);
	}

	m_auditLogQueue.front()->add_request_body_chunk(data, data_len);
}

void TcpFlow::requestEndBodyCb() {
	m_auditLogQueue.front()->end_request_body();
}

void TcpFlow::requestFinishedCb() {
	m_auditLogQueue.front()->end_request();

	if (!m_logResponses) {
		m_auditLogQueue.front()->end_transaction();
		dumpAuditLog();
	}
}

// Response handling
void TcpFlow::responseStartedCb(int status, const char *message, int message_len, int http_minor_version, time_t tv_sec, suseconds_t tv_usec) {
	waf2_log(WAF2_LOG_TRACE, "-- Response parsed:");
	waf2_log(WAF2_LOG_TRACE, "\tResponse status: %d", status);
	waf2_log(WAF2_LOG_TRACE, "\tResponse message: %.*s", message_len, message);
	waf2_log(WAF2_LOG_TRACE, "\tResponse HTTP version: 1.%d", http_minor_version);

	if (m_logResponses) {
		char log_time[64];
		snprintf(log_time, sizeof(log_time), "%lu.%lu", tv_sec, tv_usec);
		m_auditLogQueue.front()->start_response(status, std::string(message, message_len), http_minor_version, log_time);
		m_auditLogQueue.front()->start_response_hdrs();
	}

	waf2_log(WAF2_LOG_TRACE, "Headers:");
}

void TcpFlow::responseHeaderCb(const char *name, int name_len, const char *value, int value_len)  {
	waf2_log(WAF2_LOG_TRACE, "\t|%.*s: '%.*s'", name_len, name, value_len, value);

	if (m_logResponses) {
		m_auditLogQueue.front()->add_response_hdr(name, name_len, value, value_len);
	}
}

void TcpFlow::responseEndHeadersCb() {
	if (m_logResponses) {
		m_auditLogQueue.front()->end_response_hdrs();
	}

	waf2_log(WAF2_LOG_TRACE, "-------");
}

void TcpFlow::responseBodyStartCb() {
	if (m_logResponses && m_logResponseBody) {
		m_auditLogQueue.front()->start_response_body();
	}
}

void TcpFlow::responseBodyChunkCb(const char *data, int data_len) {
	if (waf2_log_level() == WAF2_LOG_TRACE) {
		waf2_log(WAF2_LOG_TRACE, "RESPONSE BODY_CHUNK:");
		util_hex_dump(data, data_len);
	}

	if (m_logResponses && m_logResponseBody) {
		m_auditLogQueue.front()->add_response_body_chunk(data, data_len);
	}
}

void TcpFlow::responseEndBodyCb() {
	if (m_logResponses && m_logResponseBody) {
		m_auditLogQueue.front()->end_response_body();
	}
}

void TcpFlow::responseFinishedCb() {
	if (m_logResponses) {
		m_auditLogQueue.front()->end_response();
		m_auditLogQueue.front()->end_transaction();
		dumpAuditLog();
	}
}

void TcpFlow::dumpAuditLog() {
	m_auditLogQueue.front()->dump_report(stdout);

	// we are finished with this request, lets remove it from the pipeline
	m_auditLogQueue.pop();
	waf2_log(WAF2_LOG_TRACE, "HTTP Pipeline finished transaction. Pipeline size now: %lu", m_auditLogQueue.size());
}
