#ifndef __TRANSACTION_LOGGER__eabda944
#define __TRANSACTION_LOGGER__eabda944

#include "yajl_gen.h"
#include <stdio.h>
#include <vector>
#include <string>
#include <stdint.h>

// Yajl generator (C++ RAII edition :)
struct Yajl {
	yajl_gen g;
	Yajl():g(yajl_gen_alloc(NULL)) {}
	~Yajl() { yajl_gen_free(g); }

	struct Map {
		yajl_gen &g;
		explicit Map(Yajl &y):g(y.g) { yajl_gen_map_open(g); }
		~Map() { yajl_gen_map_close(g); }
		void gen_null(const std::string &k) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); yajl_gen_null(g); }
		void gen_str(const std::string &k, const std::string &v) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); yajl_gen_string(g, (unsigned char *)v.data(), v.size()); }
		void gen_bool(const std::string &k, bool v) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); yajl_gen_bool(g, v); }
		void gen_integer(const std::string &k, long long int v) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); yajl_gen_integer(g, v); }
		void gen_double(const std::string &k, double v) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); yajl_gen_double(g, v); }
		void gen_key(const std::string &k) { yajl_gen_string(g, (unsigned char *)k.data(), k.size()); }
	};

	struct Array {
		yajl_gen &g;
		explicit Array(Yajl &y):g(y.g) { yajl_gen_array_open(g); }
		~Array() { yajl_gen_array_close(g); }
		void gen_null() { yajl_gen_null(g); }
		void gen_str(const std::string &v) { yajl_gen_string(g, (unsigned char *)v.data(), v.size()); }
		void gen_bool(bool v) { yajl_gen_bool(g, v); }
		void gen_integer(long long int v) { yajl_gen_integer(g, v); }
		void gen_double(double v) { yajl_gen_double(g, v); }
	};

	std::string get_json_str() const {
		const unsigned char *buf;
		size_t len;
		yajl_gen_get_buf(g, &buf, &len);
		return std::string((char*)buf, len);
	}
};

// Thin wrapper around the API (TBD: may be removed as extra level of abstraction if not needed)
class TransactionLogger {
	std::string m_conn_est_time;
	std::string m_transaction_id;
	std::string m_remote_addr;
	uint16_t m_remote_port;
	std::string m_local_addr;
	uint16_t m_local_port;
	std::string m_method;
	std::string m_uri;
	std::vector< std::pair<std::string, std::string> > m_headers;
	std::string m_body;
	std::string m_request_log_time;
	// response
	bool m_response_logged;
	int m_response_status;
	std::string m_response_reason;
	int m_response_http_version;
	std::string m_response_log_time;
	std::vector< std::pair<std::string, std::string> > m_response_headers;
	std::string m_response_body;

	TransactionLogger( const TransactionLogger& ); // non-copyable
public:
	TransactionLogger():m_remote_port(0), m_local_port(0), m_response_logged(false), m_response_status(0), m_response_http_version(0) {
	}

	~TransactionLogger() {
	}

	void start_transaction() {
		// transaction info
		m_conn_est_time = "";
		m_transaction_id = "";
		m_remote_addr = "";
		m_remote_port = 0;
		m_local_addr = "";
		m_local_port = 0;
		m_method = "";
		m_uri = "";
		m_headers.clear();
		m_body.clear();
		m_request_log_time.clear();
		// response
		m_response_logged = false;
		m_response_status = 0;
		m_response_reason.clear();
		m_response_http_version = 0;
		m_response_log_time.clear();
	}

	void emit_transaction_info(const char *conn_est_time, const char *transaction_id, const char *remote_addr, int remote_port, const char *local_addr, int local_port) {
		m_conn_est_time = conn_est_time;
		m_transaction_id = transaction_id;
		m_remote_addr = remote_addr;
		m_remote_port = remote_port;
		m_local_addr = local_addr;
		m_local_port = local_port;
	}

	// Request
	void start_request(const char *method, const char *uri, const char *request_log_time) {
		m_method = method;
		m_uri = uri;
		m_request_log_time = request_log_time;
	}

	void start_request_hdrs() {
		m_headers.clear();
	}

	void add_request_hdr(const char *name, int name_len, const char *value, int value_len) {
		m_headers.push_back(std::make_pair(std::string(name, name_len), std::string(value, value_len)));
	}

	void end_request_hdrs() {
	}

	void start_request_body() {
		m_body.clear();
	}

	void add_request_body_chunk(const char *data, int data_len) {
		m_body += std::string(data, data_len);
	}

	void end_request_body() {
	}

	void end_request() {
	}

	// Response
	void start_response(int response_status, const std::string &response_reason, int http_version, const char *response_log_time) {
		m_response_logged = true;
		m_response_status = response_status;
		m_response_reason = response_reason;
		m_response_http_version = http_version;
		m_response_log_time = response_log_time;
	}

	void start_response_hdrs() {
		m_response_headers.clear();
	}

	void add_response_hdr(const char *name, int name_len, const char *value, int value_len) {
		m_response_headers.push_back(std::make_pair(std::string(name, name_len), std::string(value, value_len)));
	}

	void end_response_hdrs() {
	}

	void start_response_body() {
		m_response_body.clear();
	}

	void add_response_body_chunk(const char *data, int data_len) {
		m_response_body += std::string(data, data_len);
	}

	void end_response_body() {
	}

	void end_response() {
	}

	void end_transaction() {
	}

	void dump_report(FILE *f) {
		Yajl y;
		{
			Yajl::Map root(y);
			root.gen_key("data");
			{
				Yajl::Map data(y);
				data.gen_key("connection");
				{
					Yajl::Map transaction(y);
					transaction.gen_str("time", m_conn_est_time);
					transaction.gen_integer("client_port", m_remote_port);
					transaction.gen_str("client_address", m_remote_addr);
					transaction.gen_integer("server_port", m_local_port);
					transaction.gen_str("server_address", m_local_addr);
				}
				data.gen_key("request");
				{
					Yajl::Map request(y);
					request.gen_str("time", m_request_log_time);
					request.gen_str("method", m_method);
					request.gen_str("uri", m_uri);
					request.gen_key("headers");
					{
						Yajl::Map headers(y);
						for (std::vector<std::pair<std::string, std::string> >::iterator it=m_headers.begin(); it!=m_headers.end(); ++it) {
							headers.gen_str(it->first, it->second);
						}
					}
					request.gen_str("body", m_body);
				}
				if (m_response_logged) {
					data.gen_key("response");
					{
						Yajl::Map response(y);
						response.gen_integer("status", m_response_status);
						response.gen_str("reason", m_response_reason);
						response.gen_integer("protocol_version", m_response_http_version);
						response.gen_str("time", m_response_log_time);
						response.gen_key("headers");
						{
							Yajl::Map headers(y);
							for (std::vector<std::pair<std::string, std::string> >::iterator it=m_response_headers.begin(); it!=m_response_headers.end(); ++it) {
								headers.gen_str(it->first, it->second);
							}
						}
						response.gen_str("body", m_response_body);
					}
				}
			}
		}

		std::string outJsonStr = y.get_json_str();
		fprintf(f, "%s\n", outJsonStr.c_str());
	}
};

#endif // TRANSACTION_LOGGER__eabda944
