#define WAF2_LOGGING_ENABLE

#include "TcpFlow.h"
#include "HtsenseUtil.h"
#include "waf2_logging.h"
#include <stdlib.h>
#include <stdio.h>
#include <nids.h>
#include <argp.h>
// TODO:: remove these?
#include "picojson.h"
#include <sstream>
#include <fstream>

// General application info
#define PROGRAM_NAME "htsense"
#define PROGRAM_VERSION "0.1"
#define PROGRAM_DESC "Record and analyze live http streams by passively listening on network intreface."
#define PROGRAM_BUG_ADDRESS "Vitaly <vitalyn@forcenock.com>"

// Global config structure
typedef struct conf_t {
	bool verbose;
	bool debug;
	bool logResponses;
	bool logResponseBody;
	std::string interface;
	std::string fname;
	std::string filter;
} conf_t;

// Global config (with defaults).
// If there are options on the command line then they may override this
static conf_t g_conf = {
	false, // verbose
	false, // debug
	false, // logResponses
	false, // logResponseBody
	"all", // interface (default, if not specified on command line)
	"", // fname
	"" // filter
};

// Command line arguments description
const char *argp_program_version = PROGRAM_NAME " " PROGRAM_VERSION;
const char *argp_program_bug_address = PROGRAM_BUG_ADDRESS;
const char *argp_program_desc = PROGRAM_NAME " - " PROGRAM_DESC;
const char *argp_program_args_desc = "<filter expression>";
static struct argp_option argp_options[] = {
	{"verbose", 'v', NULL, 0, "Produce verbose output", 0 },
	{"debug", 'd', NULL, 0, "Produce debug output", 0 },
	{"log-responses", 'r', NULL, 0, "Include responses in audit log", 0 },
	{"log-response-body", 'R', NULL, 0, "Include response bodies in audit log (only meaningful if log-responses is enabled)", 0 },
	{"streams-hashsize", 's', "num", 0, "Size of the hash table used for storing TCP connection information (a maximum of 3/4 * n_tcp_streams TCP connections will be followed simultaneously). Default value: 1024", 0 },
	{"hosts-hashsize", 'h', "num", 0, "Size of the hash table used for storing IP defragmentation information. Default value: 256", 0 },
	{"interface", 'i', "interface", 0, "Network interface name to listen on. Default value: all", 0 },
	{"file", 'f', "file", 0, ".pcap file name to parse. Either this or -i must be specified!", 0 },
	{ 0 }
};

// parse single option callback
static error_t opt_parse_cb(int key, char *arg, struct argp_state *state) {
	switch (key) {
		case 'v': g_conf.verbose = true; break;
		case 'd': waf2_log_set_level(WAF2_LOG_TRACE); break; // set log level to trace to show debug info
		case 'r': g_conf.logResponses = true; break;
		case 'R': g_conf.logResponseBody = true; break;
		case 'i': g_conf.interface = arg; break;
		case 'f': g_conf.fname = arg; break;
		case 's': if (!my_atoi(nids_params.n_tcp_streams, arg) || nids_params.n_tcp_streams <= 0) {
				waf2_log(WAF2_LOG_ERROR, "The tcp-streams command line option should specify non-zero positive integer value.");
				argp_usage(state);
				return ARGP_ERR_UNKNOWN;
			  };
			  break;
		case 'h': if (!my_atoi(nids_params.n_hosts, arg) || nids_params.n_hosts <= 0) {
				waf2_log(WAF2_LOG_ERROR, "The hosts command line option should specify non-zero positive integer value.");
				argp_usage(state);
				return ARGP_ERR_UNKNOWN;
			  };
			  break;
		case ARGP_KEY_NO_ARGS: argp_usage(state); break;
		case ARGP_KEY_ARG: g_conf.filter += g_conf.filter.empty() ? arg : std::string(" ") + arg; break;
		default:
			return ARGP_ERR_UNKNOWN;
	};

	return 0;
}

// Command line arguments descriptor
static struct argp argp = { argp_options, opt_parse_cb, argp_program_args_desc, PROGRAM_DESC };

// Callback handling assembled TCP flows
void tcp_callback(struct tcp_stream *a_tcp, void **pp_flow) {
	struct timeval *ts = &(nids_last_pcap_header->ts);

	switch(a_tcp->nids_state) {
		case NIDS_JUST_EST: { // new stream (handshake)
			// decide whether to collect this stream
			if (g_conf.logResponses) a_tcp->client.collect++; // data received by a client is needed if logging the response
			a_tcp->server.collect++; // and by a server, too

			// Create our data structure maintained per tcp flow
			TcpFlow *p_flow = new TcpFlow(a_tcp->addr, ts->tv_sec, ts->tv_usec, g_conf.logResponses, g_conf.logResponseBody);
			*pp_flow = (void*)p_flow; // attach our "TCP Flow" structure to nids flow
			waf2_log(WAF2_LOG_TRACE, "Established: %s", p_flow->name().c_str());
			break;
		}
		case NIDS_DATA: { // new data arrived
			TcpFlow *p_flow = (TcpFlow*)(*pp_flow);

			if (a_tcp->client.count_new) {
				// data server to client (responses)
				if (waf2_log_level() == WAF2_LOG_TRACE) {
					waf2_log(WAF2_LOG_TRACE, "S2C: %s:", p_flow->name().c_str());
					util_hex_dump(a_tcp->client.data, a_tcp->client.count_new);
					waf2_log(WAF2_LOG_TRACE, ">> PARSING RESPONSE DATA:");
				}

				int bytes_processed = p_flow->responseFlow.push(a_tcp->client.data, a_tcp->client.count_new, ts->tv_sec, ts->tv_usec);

				if (waf2_log_level() == WAF2_LOG_TRACE) {
					waf2_log(WAF2_LOG_TRACE, "<< RESPONSE BYTES PROCESSED: %d", bytes_processed);
					waf2_log(WAF2_LOG_TRACE, "=====================================================================");
				}

				if (bytes_processed == -1) {
					// If HTTP parser error occurred - stop collecting data on this connection (in both directions)
					if (g_conf.logResponses) a_tcp->client.collect--;
					a_tcp->server.collect--;
					waf2_log(WAF2_LOG_ERROR, "HTTP response parse error. Stop collecting data from: %s; requests_total_bytes=%lu, responses_total_bytes=%lu",
						p_flow->name().c_str(), p_flow->requestFlow.get_total_bytes(), p_flow->responseFlow.get_total_bytes());
					delete p_flow;
				}
			}
			else {
				// data client to server (requests)
				if (waf2_log_level() == WAF2_LOG_TRACE) {
					waf2_log(WAF2_LOG_TRACE, "C2S: %s:", p_flow->name().c_str()); 
					util_hex_dump(a_tcp->server.data, a_tcp->server.count_new);
					waf2_log(WAF2_LOG_TRACE, ">> PARSING REQUEST DATA:");
				}

				int bytes_processed = p_flow->requestFlow.push(a_tcp->server.data, a_tcp->server.count_new, ts->tv_sec, ts->tv_usec);

				if (waf2_log_level() == WAF2_LOG_TRACE) {
					waf2_log(WAF2_LOG_TRACE, "<< REQUEST BYTES PROCESSED: %d", bytes_processed);
					waf2_log(WAF2_LOG_TRACE, "=====================================================================");
				}

				if (bytes_processed == -1) {
					// If HTTP parser error occurred - stop collecting data on this connection (in both directions)
					if (g_conf.logResponses) a_tcp->client.collect--;
					a_tcp->server.collect--;
					waf2_log(WAF2_LOG_ERROR, "HTTP request parse error. Stop collecting data from: %s; requests_total_bytes=%lu, responses_total_bytes=%lu",
						p_flow->name().c_str(), p_flow->requestFlow.get_total_bytes(), p_flow->responseFlow.get_total_bytes());
					delete p_flow;
				}
			}

			break;
		}
		case NIDS_CLOSE:
		case NIDS_RESET:
		case NIDS_TIMED_OUT: {
			TcpFlow *p_flow = (TcpFlow*)(*pp_flow);

#ifdef WAF2_LOGGING_ENABLE
			const char *state_name = "UNKNOWN";

			switch (a_tcp->nids_state) {
				case NIDS_CLOSE: state_name = "CLOSE"; break;
				case NIDS_RESET: state_name = "RESET"; break;
				case NIDS_TIMED_OUT: state_name = "TIMED OUT"; break;
			}

			waf2_log(WAF2_LOG_TRACE, "Closed: %s [%s]: requests_total_bytes=%lu, responses_total_bytes=%lu",
				p_flow->name().c_str(), state_name, p_flow->requestFlow.get_total_bytes(), p_flow->responseFlow.get_total_bytes());
#endif

			delete p_flow;
			break;
		}
		case NIDS_EXITING: {
			waf2_log(WAF2_LOG_TRACE, "NIDS exiting");
		}
	}
}

// Tell nids not to check packet checksums.
// nids acts like this: if (src_ip & mask == netaddr) do(action);
struct nids_chksum_ctl chksum_ctl[] = {
	{
		0, // netaddr
		0, // mask
		NIDS_DONT_CHKSUM // action
	}
};

int main(int argc, char *argv[]) {
	// Initialize logger
	waf2_log_set_level(WAF2_LOG_INFO);

	// Process command line arguments
	argp_parse(&argp, argc, argv, 0, 0, &g_conf);

	if (g_conf.interface != "all" && !g_conf.fname.empty()) {
		fprintf(stderr, "ERROR: You can't specify both .pcap file name (-f switch) and interface name (-i switch) at the same fime.\nUse either one of these, not both!\n");
		exit(-1);
	}

	// Initialize NIDS global parameters

	if (!g_conf.fname.empty()) {
		// Read requests from file
		fprintf(stderr, "Reading from pcap file '%s'\n", g_conf.fname.c_str());
		nids_params.device = NULL;
		nids_params.filename = (char*)g_conf.fname.c_str();
	}
	else {
		fprintf(stderr, "Listening to live traffic on interface: '%s'\n", g_conf.interface.c_str());
		// net interface from which to collect data
		nids_params.device = (char *)g_conf.interface.c_str();
		// Put the interface in promiscuous mode
		nids_params.promisc = 1;
		// Timeout (ms) passed to pcap_open_live()
		nids_params.pcap_timeout = 2048; // default = 1024ms
	}

	// pcap filter (similar to tcpdump)
	nids_params.pcap_filter = (char*)g_conf.filter.c_str();

	waf2_log(WAF2_LOG_INFO, PROGRAM_NAME " listening on interface '%s' with filter '%s'...", g_conf.interface.c_str(), g_conf.filter.c_str());

	nids_register_chksum_ctl(chksum_ctl, 1);
	// Attach to pcap and prepare to listen on given interface
	if (!nids_init()) {
		waf2_log(WAF2_LOG_ERROR, "NIDS/PCAP initialization failed: %s", nids_errbuf);
		exit(-1);
	}

	nids_register_tcp((void*)tcp_callback);

	// Start listening for packets and reassembling flows
	// note: normally, nids_run() call will never return.
	nids_run();
	return 0;
}
